from typing import List

from pyspark.sql import SparkSession
from pyspark.sql.functions import col, lit
from pyspark.sql.types import (
    BooleanType,
    DateType,
    DoubleType,
    FloatType,
    IntegerType,
    LongType,
    StringType,
    TimestampType,
)

spark = SparkSession.builder.getOrCreate()


# Check table existence
def table_exists(table_path: str) -> bool:
    hconf = spark._jsc.hadoopConfiguration()

    # get the Hadoop FileSystem bound to that config
    fs = spark.sparkContext._jvm.org.apache.hadoop.fs.FileSystem.get(hconf)

    # build a Hadoop Path for your ABFSS location
    path = spark.sparkContext._jvm.org.apache.hadoop.fs.Path(f"{table_path}")

    # ask the FileSystem if it’s there
    return fs.exists(path)


# JDBC Postgres reader
def query_postgres_table(
    hostname: str, database: str, table: str, user: str, password: str, port: int = 5432
):
    jdbc_url = f"jdbc:postgresql://{hostname}:{port}/{database}?sslmode=require"
    connection_props = {
        "user": user,
        "password": password,
        "driver": "org.postgresql.Driver",
    }
    return (
        spark.read.format("jdbc")
        .option("url", jdbc_url)
        .option("dbtable", f"public.{table}")
        .options(**connection_props)
        .load()
    )


# Map Spark types to SQL types
def spark_type_to_sql(data_type):
    if isinstance(data_type, StringType):
        return "STRING"
    if isinstance(data_type, IntegerType):
        return "INT"
    if isinstance(data_type, LongType):
        return "BIGINT"
    if isinstance(data_type, DoubleType):
        return "DOUBLE"
    if isinstance(data_type, FloatType):
        return "FLOAT"
    if isinstance(data_type, BooleanType):
        return "BOOLEAN"
    if isinstance(data_type, TimestampType):
        return "TIMESTAMP"
    if isinstance(data_type, DateType):
        return "DATE"
    return data_type.simpleString().upper()


# align dataframe to Delta table schema
def align_df_to_delta_schema(df, table_path: str):
    target_schema = spark.read.format("delta").load(table_path).schema
    df_cols = set(df.columns)
    aligned = df

    for f in target_schema:
        if f.name not in df_cols:
            aligned = aligned.withColumn(f.name, lit(None).cast(f.dataType))

    return aligned.select([f.name for f in target_schema])


# Create/update via DataFrame schema
def create_or_update_table_from_df_schema(df, table_path: str):
    if not table_exists(table_path):
        df.limit(0).write.format("delta").mode("overwrite").option("path", table_path).save()
        return

    existing = spark.read.format("delta").load(table_path).schema
    existing_fields = {f.name for f in existing.fields}
    additions = [f for f in df.schema.fields if f.name not in existing_fields]
    if not additions:
        print("No schema changes")
        return

    for fld in additions:
        sql_type = spark_type_to_sql(fld.dataType)
        null_clause = "" if fld.nullable else "NOT NULL"
        spark.sql(
            f"ALTER TABLE delta.`{table_path}` ADD COLUMNS ({fld.name} {sql_type} {null_clause})"
        )
        print(f"Added column {fld.name}")


# Merge via SQL
def merge_df_to_table(source_df, target_table_path: str, pk_cols: List[str] | str):
    """
    Upsert (MERGE) source_df into target_table_path deduplicating by pk_cols.

    - source_df: input DataFrame
    - target_table_path: Full abfs path for target Delta table
    - pk_cols: single column name or list of column names representing the key
    """
    # Normalize pk_cols to a list
    if isinstance(pk_cols, str):
        pk_list = [pk_cols]
    else:
        pk_list = pk_cols

    # Deduplicate source on the key columns
    deduped = source_df.dropDuplicates(pk_list)

    # Register staging view
    deduped.createOrReplaceTempView("_stg")

    # Build the ON clause for composite keys
    on_clause = " AND ".join(f"target.{c} = src.{c}" for c in pk_list)

    # Execute the MERGE
    merge_sql = f"""
      MERGE INTO delta.`{target_table_path}` AS target
      USING _stg AS src
        ON {on_clause}
      WHEN MATCHED THEN
        UPDATE SET *
      WHEN NOT MATCHED THEN
        INSERT *
    """
    spark.sql(merge_sql)
    print(f"Merged into delta.`{target_table_path}` deduplicated on {pk_list}")


# Watermark helpers
def get_watermark(wm_table_path: str, table_name: str) -> int:
    df = spark.read.format("delta").load(wm_table_path).filter(col("table_name") == table_name)
    return df.collect()[0]["ts"] if df.count() else 0


def update_watermark(wm_table_path: str, table_name: str, ts: int):
    wm_df = spark.createDataFrame([(table_name, ts)], ["table_name", "ts"])
    wm_df.createOrReplaceTempView("new_wm")
    spark.sql(
        f"""
      MERGE INTO delta.`{wm_table_path}` AS target
      USING new_wm AS src
        ON target.table_name = src.table_name
      WHEN MATCHED THEN UPDATE SET ts = src.ts
      WHEN NOT MATCHED THEN INSERT (table_name, ts) VALUES (src.table_name, src.ts)
    """
    )
    print(f"Watermark for {table_name} set to {ts}")



import aiohttp

import logging
from typing import Any
from pydantic import BaseModel

logger = logging.getLogger(__name__)


class HttpResponse(BaseModel):
    """Model for HTTP response."""

    success: bool
    status: int
    headers: dict
    content: Any

    class config:
        from_attributes = True


async def get_async(
    url: str,
    accessToken: str = None,  # type: ignore
    verifySSL: bool = True,
    params: dict = None,  # type: ignore
    headers: dict = None,  # type: ignore
) -> HttpResponse:
    """Send an asynchronous GET request."""
    logger.debug(f"get_async: URL: {url}")
    try:
        async with aiohttp.ClientSession() as session:
            if not headers:
                headers = {
                    "Accept": "application/json",
                }

            if accessToken:
                headers["Authorization"] = f"Bearer {accessToken}"

            async with session.get(
                url, headers=headers, params=params, ssl=verifySSL
            ) as response:

                http_response = HttpResponse(
                    success=response.status in range(200, 300),
                    status=response.status,
                    headers=dict(response.headers),
                    content=await response.json() if response.content_type == "application/json" else await response.text(),
                )

                return http_response
    except aiohttp.ClientResponseError as e:
        logger.error(f"async get failed with status code {e.status}. {e}")
        raise
    except Exception as e:
        logger.error(f"async get failed with exception {e}")
        raise


async def post_async(
    url: str,
    accessToken: str = None,  # type: ignore
    verifySSL: bool = True,
    params: dict = None,  # type: ignore
    data: Any = None,  # type: ignore
    json: Any = None,  # type: ignore
    headers: dict = None,  # type: ignore
) -> HttpResponse:
    """Send an asynchronous POST request."""
    try:
        if json and data:
            raise ValueError("Cannot specify both 'json' and 'data' parameters.")

        logger.debug(f"post_async: URL: {url}")

        async with aiohttp.ClientSession() as session:
            if not headers:
                headers = {
                    "Accept": "application/json",
                }

            if accessToken:
                headers["Authorization"] = f"Bearer {accessToken}"

            async with session.post(
                url,
                headers=headers,
                params=params,
                data=data,
                json=json,
                ssl=verifySSL,
            ) as response:

                http_response = HttpResponse(
                    success=response.status in range(200, 300),
                    status=response.status,
                    headers=dict(response.headers),
                    content=await response.json() if response.content_type == "application/json" else await response.text(),
                )

                return http_response
    except aiohttp.ClientResponseError as e:
        logger.error(f"async post failed with status code {e.status}. {e}")
        raise
    except Exception as e:
        logger.error(f"async post failed with exception {e}")
        raise



from .azure_ai_language import (  # noqa F401
    detect_language,
    detect_languages_bulk,
    get_key_phrases,
    get_key_phrases_bulk,
    get_sentiment_analysis,
    get_sentiment_analysis_bulk,
)
from .data_services import (  # noqa F401
    align_df_to_delta_schema,
    create_or_update_table_from_df_schema,
    get_watermark,
    merge_df_to_table,
    query_postgres_table,
    spark_type_to_sql,
    table_exists,
    update_watermark,
)



import aiohttp
import asyncio
import json
from typing import Any, Dict, List, Optional
from aiohttp.client_exceptions import ClientResponseError

from .async_http_client import get_async, post_async
from .azure_ai_language_job_responses import KeyPhraseJobResponse, SentimentAnalysisJobResponse


async def _post_text_analytics(
    endpoint: str, key: str, payload: dict, max_retries: int = 10, backoff: float = 1.0
) -> Optional[Dict[str, Any]]:
    """
    Generic call to the Analyze-Text API.
    """
    url = f"{endpoint}/language/:analyze-text?api-version=2024-11-01"
    headers = {
        "Content-Type": "application/json",
        "Ocp-Apim-Subscription-Key": key,
    }

    for attempt in range(1, max_retries + 1):
        try:
            resp = await post_async(url, json=payload, headers=headers)
            if resp.success:
                return resp.content
            else:
                return None
        except ClientResponseError as e:
            if e.status == 429:  # Too Many Requests
                if attempt < max_retries:
                    print(
                        f"[ERROR] __post_text_analytics failed (attempt {attempt+1}/{max_retries}): {e}"
                    )
                    await asyncio.sleep(backoff * (2 ** (attempt - 1)))
                else:
                    # all retries failed
                    raise
            else:
                # Raise other exceptions immediately
                raise
        except Exception:
            raise


async def _post_text_analytics_job(
    endpoint: str, key: str, payload: dict, max_retries: int = 10, backoff: float = 1.0
) -> Optional[str]:
    """
    Generic call to the Analyze-Text API. Returns the operation location header.
    """
    url = f"{endpoint}/language/analyze-text/jobs?api-version=2024-11-01"
    headers = {
        "Content-Type": "application/json",
        "Content-Length": str(len(json.dumps(payload))),
        "Ocp-Apim-Subscription-Key": key,
    }

    for attempt in range(1, max_retries + 1):
        try:
            resp = await post_async(url, json=payload, headers=headers)
            if resp.success:
                return resp.headers.get("operation-location", "")
            else:
                return None
        except ClientResponseError as e:
            if e.status == 429:  # Too Many Requests
                if attempt < max_retries:
                    print(
                        f"[ERROR] __post_text_analytics_job failed (attempt {attempt+1}/{max_retries}): {e}"
                    )
                    await asyncio.sleep(backoff * (2 ** (attempt - 1)))
                else:
                    # all retries failed
                    raise
            else:
                # Raise other exceptions immediately
                raise
        except Exception:
            raise


async def _get_text_analytics_job_result(
    operation_location: str,
    key: str,
    max_retries: int = 10,
    backoff: float = 1.0,
) -> Optional[Dict[str, Any]]:
    """
    Get the result of a text analytics job.
    """
    headers = {
        "Content-Type": "application/json",
        "Ocp-Apim-Subscription-Key": key,
    }

    last_exc = None
    for attempt in range(1, max_retries + 1):
        try:
            resp = await get_async(operation_location, headers=headers)
            if resp.success:
                return resp.content
            else:
                return None
        except Exception as ex:
            last_exc = ex
            if attempt < max_retries:
                await asyncio.sleep(backoff * (2 ** (attempt - 1)))
            else:
                # all retries failed
                raise RuntimeError(
                    f"get_text_analytics_job_result failed after {max_retries} attempts"
                ) from last_exc


async def detect_language(endpoint: str, key: str, id: str, text: str) -> Dict[str, str]:
    """
    Call the Analyze-Text LanguageDetection task
    """
    try:
        payload = {
            "kind": "LanguageDetection",
            "parameters": {"modelVersion": "latest"},
            "analysisInput": {"documents": [{"id": id, "text": text[:5120]}]},
        }

        resp = await _post_text_analytics(endpoint, key, payload)
        if not resp:
            return {
                "id": id,
                "language": "en",  # default to english if no response
            }

        lang = resp["results"]["documents"][0]["detectedLanguage"]["iso6391Name"]
        return {
            "id": id,
            "detectedLanguage": lang,
        }
    except Exception as e:
        print(f"[ERROR] detect_language failed for id {id}: {e}")
        raise


async def detect_languages_bulk(
    endpoint: str,
    key: str,
    conversations: List[Dict[str, str]],
    id_column_name: str = "conversation_id",
    text_column_name: str = "messages",
) -> List[Dict[str, str]]:
    async with aiohttp.ClientSession():
        tasks = [
            detect_language(
                endpoint,
                key,
                row[id_column_name],
                row[text_column_name],
            )
            for row in conversations
        ]
        return await asyncio.gather(*tasks)


async def get_key_phrases(
    endpoint: str,
    key: str,
    id: str,
    text: str,
    language: str,
) -> Dict[str, Any]:
    """
    Call the Analyze-Text KeyPhraseExtraction task.
    """
    try:
        payload = {
            "kind": "KeyPhraseExtraction",
            "analysisInput": {"documents": [{"id": id, "language": language, "text": text}]},
            "tasks": [
                {
                    "kind": "KeyPhraseExtraction",
                    "parameters": {"modelVersion": "latest"},
                }
            ]
        }

        operation_location = await _post_text_analytics_job(endpoint, key, payload)
        if not operation_location:
            return {
                "id": id,
                "keyPhrases": [],
            }

        results = None
        while True:
            job_result = await _get_text_analytics_job_result(
                operation_location=operation_location,
                key=key,
            )
            if not job_result:
                raise RuntimeError(
                    f"get_key_phrases failed to get job result for id {id}"
                )

            job_response = KeyPhraseJobResponse(**job_result)
            # print(f"get_key_phrases: job {job_response.jobId} status: {job_response.status}")

            if job_response.status in ["notStarted", "running"]:
                await asyncio.sleep(5)
                continue

            if job_response.status == "succeeded":
                results = job_response.tasks.items[0].results
                break
            elif job_response.status == "failed":
                raise RuntimeError(
                    f"get_key_phrases failed for id {id}: {job_response.errors}"
                )

        if not results:
            return {
                "id": id,
                "keyPhrases": [],
            }

        return {
            "id": id,
            "keyPhrases": results.documents[0].keyPhrases if results.documents else [],
        }
    except Exception as e:
        print(f"[ERROR] get_key_phrases: {e}")
        raise


async def get_key_phrases_bulk(
    endpoint: str,
    key: str,
    conversations: List[Dict[str, str]],
    id_column_name: str = "conversation_id",
    text_column_name: str = "messages",
    language_column_name: str = "detected_language",
) -> List[Dict[str, Any]]:
    async with aiohttp.ClientSession():
        tasks = [
            get_key_phrases(
                endpoint=endpoint,
                key=key,
                id=row[id_column_name],
                text=row[text_column_name],
                language=row[language_column_name],
            )
            for row in conversations
        ]
        return await asyncio.gather(*tasks)


async def get_sentiment_analysis(
    endpoint: str,
    key: str,
    id: str,
    text: str,
    language: str,
) -> Dict[str, Any]:
    """
    Call the Analyze-Text SentimentAnalysis task.
    """
    try:
        payload = {
            "kind": "SentimentAnalysis",
            "analysisInput": {"documents": [{"id": id, "language": language, "text": text}]},
            "tasks": [
                {
                    "kind": "SentimentAnalysis",
                    "parameters": {"modelVersion": "latest"},
                }
            ]
        }

        operation_location = await _post_text_analytics_job(endpoint, key, payload)
        if not operation_location:
            return {
                "id": id,
                "positive": None,
                "neutral": None,
                "negative": None,
            }

        results = None
        while True:
            job_result = await _get_text_analytics_job_result(
                operation_location=operation_location,
                key=key,
            )
            if not job_result:
                raise RuntimeError(
                    f"get_sentiment_analysis failed to get job result for id {id}"
                )

            job_response = SentimentAnalysisJobResponse(**job_result)
            # print(f"get_sentiment_analysis: job {job_response.jobId} status: {job_response.status}")

            if job_response.status in ["notStarted", "running"]:
                await asyncio.sleep(5)
                continue

            if job_response.status == "succeeded":
                results = job_response.tasks.items[0].results
                break
            elif job_response.status == "failed":
                raise RuntimeError(
                    f"get_key_phrases failed for id {id}: {job_response.errors}"
                )

        if not results:
            return {
                "id": id,
                "positive": None,
                "neutral": None,
                "negative": None,
            }

        scores = results.documents[0].confidenceScores
        return {
            "id": id,
            "positive": scores.positive,
            "neutral": scores.neutral,
            "negative": scores.negative,
        }
    except Exception as e:
        print(f"[ERROR] get_sentiment_analysis id {id}: {e}")
        raise


async def get_sentiment_analysis_bulk(
    endpoint: str,
    key: str,
    conversations: List[Dict[str, str]],
    id_column_name: str = "conversation_id",
    text_column_name: str = "messages",
    language_column_name: str = "detected_language",
) -> List[Dict[str, Any]]:
    async with aiohttp.ClientSession():
        tasks = [
            get_sentiment_analysis(
                endpoint=endpoint,
                key=key,
                id=row[id_column_name],
                text=row[text_column_name],
                language=row[language_column_name],
            )
            for row in conversations
        ]
        return await asyncio.gather(*tasks)
